# Copyright (c) 2006, Determina Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of Determina Inc. nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

require 'timeout'

#
# VMware vmrun Ruby wrapper
#
# by Alexander Sotirov <asotirov@determina.com>
#

module VMware

#
# The VMrun class provides an API for controlling VMware Workstation 5.5
# through the vmrun.exe command line interface
#

class VMrun

  #
  # Configuration
  #

  # Path to the VMrun binary

  VMRUN_PATH = 'c:/Program Files/VMware/VMware Workstation/vmrun.exe'

  # Timeout for vmrun.exe, in seconds

  VMRUN_TIMEOUT = 2*60

  # Enable logging out vmrun.exe output

  VMRUN_LOGGING = false
 
  # Supported versions of vmrun.exe

  VMRUN_VERSIONS = [
    '5.5.1 build-19175',
    '5.5.2 build-29772',
    '5.5.3 build-34685'
  ]

private

  #
  # Logging
  #

  def log(msg)
    $stdout.print(msg + "\n")
    $stdout.flush
  end

  #
  # Executes a block with a timeout specified in seconds. Nested timeouts are
  # supported. On timeout raises a Timeout::Error exception.
  #
  # Options:
  #
  #   seconds  Timeout in seconds
  #

  def timeout(timeout)

    # Create an anonymous exception class to support nested timeouts
    e = Class.new(Exception)

    begin
      Timeout::timeout(timeout, e) {
        yield
      }
    rescue e
      raise Timeout::Error
    end
  end

  #
  # Executes a command on the local system with a timeout. Logs the output and
  # returns it as an array of lines. On timeout kills the command and raises a
  # Timeout::Error exception.
  #
  # Options:
  #
  #   path     Path to executable
  #   args     Arguments array
  #   timeout  Timeout in seconds (set to 0 to disable)
  #   logging  Enables output logging (defaults to true)
  #

  def execute(path, args=[], timeout=0, logging=true)

    output = []

    log("Running #{path} #{args.join(' ')}") if logging

    pstdout = IO::pipe  # pipe[0] for read, pipe[1] for write
    pstderr = IO::pipe

    pid = fork {
      # Create a new process group
      Process.setsid

      # Connect stdin and stdout to the pipes
      pstdout[0].close
      STDOUT.reopen(pstdout[1])
      pstdout[1].close

      pstderr[0].close
      STDERR.reopen(pstderr[1])
      pstderr[1].close

      # Execute the command
      exec([path, path], *args)
    }

    # Close the write ends of the pipes
    pstdout[1].close
    pstderr[1].close

    start = Time.now
    line = ''

    # Read the output line by line and write it to the log
    begin
      while true
        if (pipes = IO.select([pstdout[0], pstderr[0]], [], [], 1))
          pipes[0].each { |pipe|
            line = line + pipe.read_nonblock(1024) if !pipe.eof?
          }

          # Add the new line to the output array
          while match = /(.*?)\r?\n(.*)/m.match(line)
            output.push match[1]
            log(' * ' + match[1]) if logging

            line = match[2]
          end

          # If both pipes are eof, the program has terminated
          if pipes[0].length == 2 && pipes[0][0].eof? && pipes[0][1].eof?

            # Add the last line to the output
            if (line != '')
              output.push line
              log(' * ' + line) if logging
            end

            # Exit the loop
            break
          end
        end

        raise Timeout::Error if timeout != 0 && (Time.now - start > timeout)
      end

    rescue Timeout::Error
      # Add the last line to the output
      if (line != '')
        output.push line
        log(' * ' + line) if logging
      end

      # Get the process group of the child process
      pgid = Process.getpgid(pid)

      log("Timeout while executing \"#{path} #{args.join(' ')}\", sending sigterm to #{pgid}")

      # Send SIGTERM to the process group
      begin
        Process.kill(-15, pgid)
      rescue Errno::ESRCH
      end

      raise

    ensure
      pstdout[0].close if not pstdout[0].closed? 
      pstderr[0].close if not pstderr[0].closed? 

      begin
        Process.wait
      rescue Errno::ECHILD
      end
    end

    return output
  end

  #
  # Executes vmrun.exe and checks for errors. The vmrun.exe output is returned
  # as an array of lines with the newline characters removed. The array could
  # be empty.
  #

  def vmrun(cmd)
    output = execute(VMRUN_PATH, cmd, VMRUN_TIMEOUT, VMRUN_LOGGING)

    if not output.empty?
      match = /^Error: (.*)$/.match(output[0])
      if (match)
        raise "VMrun error: #{match[1]}"
      end
    end

    return output
  end
  
public

  #
  # Initializes a VMrun object and verifies that vmrun.exe exists and is a
  # supported version
  #

  def initialize
    output = vmrun([])
    if output.empty?
      raise "Error running #{VMRUN_PATH}"
    end

    match = /^vmrun version (.*?)$/.match(output[0])
    if !match
      raise "Unsupported VMware version"
    end

    if not VMRUN_VERSIONS.include?(match[1])
      raise "Unsupported VMware version #{match[1]}\n"
    end
  end

  #
  # Returns a list of the .vmx files for all running VMs
  #

  def running

    output = vmrun(['list'])
    if output.empty?
      raise "Error running #{VMRUN_PATH}"
    end

    line = output.shift

    # Get the number of running VMs
    match = /^Total running VMs: (\d+)$/.match(line)
    if !match
      raise "Cannot parse vmrun output"
    end

    num = match[1].to_i

    if (num != output.length)
      raise "VMrun list returned #{output.length} entries, #{num} expected"
    end
      
    return output
  end

  #
  # Returns true if a VM is running. The vmx argument is a path to a .vmx file.
  #

  def running?(vmx)
    return File.exists?(vmx + '.lck')
  end

  #
  # Returns a list of all snapshots of a VM
  #

  def snapshots(vmx)
    log "Getting a list of snapshots of VM #{vmx}"

    output = vmrun(['listSnapshots', vmx])
    if output.empty?
      raise "Error running #{VMRUN_PATH}"
    end

    line = output.shift

    # Get the number of snapshots
    match = /^Total snapshots: (\d+)$/.match(line)
    if !match
      raise "Cannot parse vmrun output"
    end

    num = match[1].to_i

    if (num != output.length)
      raise "VMrun list returned #{output.length} entries, #{num} expected"
    end
      
    return output
  end

  #
  # Stops a VM
  #

  def stop(vmx)
    log "Stopping VM #{vmx}"

    output = vmrun(['stop', vmx])
    if not output.empty?
      raise "Stopping #{vmx} failed: " + output.inspect
    end

    if running?(vmx)
      raise "VM #{vmx} is still running"
    end
  end
  
  #
  # Revert a VM to a snapshot
  #

  def revert(vmx, snapshot)
    log "Reverting VM #{vmx} to snapshot #{snapshot}"

    output = vmrun(['revertToSnapshot', vmx, snapshot])
    if not output.empty?
      raise "Reverting #{vmx} to #{snapshot} failed: " + output.inspect
    end
  end

  #
  # Starts a VM
  #

  def start(vmx)
    log "Starting VM #{vmx}"

    if running?(vmx)
      raise "VM #{vmx} is already running"
    end

    # start the VM with a timeout

    output = vmrun(['start', vmx])
    if not output.empty?
      raise "Starting #{vmx} failed: " +  output.inspect
    end

    # wait for it to appear in the list of running VMs
    timeout(VMRUN_TIMEOUT) {
      while not running?(vmx)
        sleep 1
      end
    }
  end
  
  #
  # Takes a snapshot
  #

  def snapshot(vmx, snapshot)

    existing = snapshots(vmx)

    if existing.include?(snapshot)
      raise "Snapshot #{snapshot} already exists"
    end

    log "Creating snapshot #{snapshot} of VM #{vmx}"

    output = vmrun(['snapshot', vmx, snapshot])
    if not output.empty?
      raise "Taking snapshot #{snapshot} of VM #{vmx} failed"
    end
  end

end 
end
